;-----------------------------------------------------------------------------
;
;						XMM Science Analysis System (SAS)
;						(c) 1997-2010 European Space Agency
;
;-----------------------------------------------------------------------------
;
;	xmmtrack --- 3-D visualization of XMM orbit and attitude data
;
;	DESCRIPTION:
;		xmmtrack is an IDL application that visualizes the S/C orbit and
;		attitude data in an ODF by means of a 3-D animation sequence
;		using IDL. The code inquires various parameters from the OAL
;		via the CALL_EXTERNAL technique (see IDL manual for detail).
;		Parameters that are accessed:
;			+ ODF revolution number
;			+ observation start/end times
;			+ actual/predicted S/C position at given time
;			+ actual S/C attitude at given time
;
;		Upon startup a scene with the Earth, the total trajectory of
;		the S/C during the observation and the actual S/C position at
;		the start is rendered and displayed in a view frame. Then an
;		animation loop is entered which propagates the satellite through
;		its pre-determined trajectory. The user is offered various
;		possibilities to alter the scene interactively, including:
;			+ change the size of the satellite
;			+ zoom in/out
;			+ change to orientation of the scence with the mouse
;			+ display entire prdicted orbit (at arbitrary epoch)
;		For more information, consult the documentation in
;		$SAS_DIR/packages/tools/doc.
;
;	CHANGE HISTORY:
;		UL, 2001-01-08: 1.0.0: - initial version (loosely based on IDL demo
;								 application d_orbit)
;		UL, 2001-12-04: 1.0.1: - siderial -> sidereal
;
;	AUTHOR:
;		U. Lammers, Mon Jan  8 16:42:46 MET 2001
;
;	REFERENCES:
;		[1]: Software Specification Document for the XMM Science Analysis
;			 Subsystem, M. Dahlem et. al., 
;
;	EXTERNAL IDL ROUTINES USED:
;		- orb__define.pro	: create an orb object
;		- trackball__define : create the trackball object
;
;	$Header: /usr/local/cvs/sasdev/task_devel/packages/tools/src/xmmtrack/xmmtrack.pro,v 1.5 2001/12/04 17:10:35 ulammers Exp $

;--------------------------------------------------------------------------
;	inquire revolution number
;--------------------------------------------------------------------------
function xmmRevolutionNr
	RETURN, CALL_EXTERNAL('libxmmtrackinfo.so', 'revolution')
end


;--------------------------------------------------------------------------
;	inquire S/C attitude at specific time using OAL
;--------------------------------------------------------------------------
function xmmAttitude, $
    time         ; IN: time in sec passed the epoch

    raDecApos = DBLARR(3)
	if (CALL_EXTERNAL('libxmmtrackinfo.so', 'xmmAttitude', time, raDecApos) $
		ne 0) then raDecApos = [0, 0, 0]
;	raDecApos = [!PI/2, -!PI/4, -90.*!PI/180.]
;	raDecApos = [0., 0., 0.]
;	print, raDecApos

	;
	;	raDecApos is three-element vector with Right Acension/Declination
	;	of star tracker viewing direction + astronomical position angle;
	;	the below constructs direction cosine matrix from intepreting these
	;	as 3-2-1 Euler angles; for this to work one needs to negate the
	;	declination+APOS
	;
	sinPhi   = sin(raDecApos[0])
	cosPhi   = cos(raDecApos[0])
	sinTheta = sin(-raDecApos[1])
	cosTheta = cos(-raDecApos[1])
    sinPsi   = sin(-raDecApos[2])
    cosPsi   = cos(-raDecApos[2])

	A11 = cosTheta*cosPhi
	A12 = cosTheta*sinPhi
    A13 = -sinTheta
    A21 = -cosPsi*sinPhi + sinPsi*sinTheta*cosPhi
    A22 = cosPsi*cosPhi + sinPsi*sinTheta*sinPhi
	A23 = sinPsi*cosTheta
    A31 = sinPsi*sinPhi + cosPsi*sinTheta*cosPhi
    A32 = -sinPsi*cosPhi + cosPsi*sinTheta*sinPhi
    A33 = cosPsi*cosTheta

    phi = acos(.5*(A11+A22+A33-1.))
    twoSinPhi = 2.*sin(phi)

	if (twoSinPhi eq 0.) then $
       twoSinPhi = twoSinPhi + 1.e-7

	angArr = FLTARR(7)
    angArr = [raDecApos[0], raDecApos[1], raDecApos[2], $
			  (A23-A32)/twoSinPhi, (A31-A13)/twoSinPhi, (A12-A21)/twoSinPhi, $
			  phi]
;	print, angArr

	RETURN, angArr
end


;--------------------------------------------------------------------------
;	Inquire S/C position in earth-centered J2000 reference frame at
;	given point in time (MET tag)
;--------------------------------------------------------------------------
function xmmPosition, $
    time       ; IN:  MET time tag

    xf = DBLARR(3)
	if (CALL_EXTERNAL('libxmmtrackinfo.so', 'xmmPosition', time, xf) ne 0) $
		then xf = [0, 0, 0]
;	print, xf
;	xf = [10000, 0, 0]
    RETURN, xf
end


;--------------------------------------------------------------------------
;	Inquire S/C position along PREDICTED orbit
;--------------------------------------------------------------------------
function xmmPredictedPosition, $
    time       ; IN:  MET tag

    xf = DBLARR(3)
	res = CALL_EXTERNAL('libxmmtrackinfo.so', 'xmmPredictedPosition', time, xf)
;	print, xf, sqrt(xf[0]*xf[0]+xf[1]*xf[1]+xf[2]*xf[2])
    RETURN, xf
end

;--------------------------------------------------------------------------
;	Construct cylinder with specified half height/radius
;--------------------------------------------------------------------------
function cylinder, $
	halfHeight,	$	;	IN: half height of cylinder
	R				;	IN: radius

	nSegments = 10
	deltaPhi = 360./FLOAT(nSegments)
	x = FLTARR(3, 2*nSegments)
	index = INTARR(nSegments+1, nSegments+2)
	for i = 0, nSegments-1 do begin
		x[0, 2*i] = halfHeight
		x[0, 2*i+1] = -halfHeight

		y = R*cos(i*deltaPhi*!PI/180.)
		x[1, 2*i] = y
		x[1, 2*i+1] = y

		z = R*sin(i*deltaPhi*!PI/180.)
		x[2, 2*i] = z
		x[2, 2*i+1] = z

		index[0:4, i] = [4, 2*i, 2*i+1, 2*i+3, 2*i+2 ]
		index[i+1, nSegments] = 2*i
		index[i+1, nSegments+1] = 2*i+1
	end
	index[3:4, nSegments-1] = [1, 0]
	index[0, nSegments] = nSegments
	index[0, nSegments+1] = nSegments

	RETURN, OBJ_NEW('IDLgrPolygon', x, POLYGONS=index, COLOR=[255, 255, 0], $
		SHADING=0, STYLE=2)
end


;---------------------------------------------------------------------------
;	Create XMM model
;---------------------------------------------------------------------------
function xmmCreateModel

    mainModel = OBJ_NEW('IDLgrModel')

	;
    ;  Create bloc1 : main satellite body in yellow.
    ;
	mainModel->Add, cylinder(.4, .07)

	;
    ;  Create bloc2 : solar array panel.
    ;
    xp=[0.20, 0.60, 0.60, 0.20, $
        0.20, 0.60, 0.60, 0.20]
    yp=[0.40,0.40, -0.00, -0.00, $
        0.40,0.40, -0.00, -0.00]
    zp=[ 0.02, 0.02, 0.02, 0.02, $
        -0.02,-0.02,-0.02,-0.02]
    bloc1Mesh = [ [4,0,1,2,3], $
                  [4,1,5,6,2], $
                  [4,4,7,6,5], $
                  [4,0,3,7,4], $
                  [4,3,2,6,7], $
                  [4,0,4,5,1] ]
    bloc2Vertices = fltarr(3,8)
    bloc2Vertices = [ [yp], [xp], [zp] ]
    bloc2Vertices = TRANSPOSE(bloc2Vertices)
    bloc2Mesh = [ [4,0,1,2,3], $
                  [4,1,5,6,2], $
                  [4,4,7,6,5], $
                  [4,0,3,7,4], $
                  [4,3,2,6,7], $
                  [4,0,4,5,1] ]
    bloc2List = FLTARR(3,24)
    bloc2NewMesh = bloc1Mesh
    j = 0
    for i = 0, 5 do begin
        bloc2List[0:2, i*4+0] = bloc2Vertices[0:2, bloc2Mesh[i*5+1]]
        bloc2List[0:2, i*4+1] = bloc2Vertices[0:2, bloc2Mesh[i*5+2]]
        bloc2List[0:2, i*4+2] = bloc2Vertices[0:2, bloc2Mesh[i*5+3]]
        bloc2List[0:2, i*4+3] = bloc2Vertices[0:2, bloc2Mesh[i*5+4]]
        bloc2NewMesh[*,i] = [4, j+0, j+1, j+2, j+3]
        j = j + 4
    endfor
    oBloc2 = OBJ_NEW('IDLgrPolygon', bloc2List, $
        POLYGONS=bloc2NewMesh, COLOR=[100,100,255] )
    mainModel->Add, oBloc2

	;
    ;  Create bloc3 : the other solar array panel.
    ;
    xp=[-0.20, -0.60, -0.60, -0.20, $
        -0.20, -0.60, -0.60, -0.20]
    yp=[0.40,0.40, -0.00, -0.00, $
        0.40,0.40, -0.00, -0.00]
    zp=[ 0.02, 0.02, 0.02, 0.02, $
        -0.02,-0.02,-0.02,-0.02]
    bloc3Vertices = fltarr(3,8)
    bloc3Vertices = [ [yp], [xp], [zp] ]
    bloc3Vertices = TRANSPOSE(bloc3Vertices)
    bloc3Mesh = [ [4,0,1,2,3], $
                  [4,1,5,6,2], $
                  [4,4,7,6,5], $
                  [4,0,3,7,4], $
                  [4,3,2,6,7], $
                  [4,0,4,5,1] ]
    bloc3List = FLTARR(3,24)
    bloc3NewMesh = bloc3Mesh
    j = 0
    for i = 0, 5 do begin
        bloc3List[0:2, i*4+0] = bloc3Vertices[0:2, bloc3Mesh[i*5+1]]
        bloc3List[0:2, i*4+1] = bloc3Vertices[0:2, bloc3Mesh[i*5+2]]
        bloc3List[0:2, i*4+2] = bloc3Vertices[0:2, bloc3Mesh[i*5+3]]
        bloc3List[0:2, i*4+3] = bloc3Vertices[0:2, bloc3Mesh[i*5+4]]
        bloc3NewMesh[*,i] = [4, j+0, j+1, j+2, j+3]
        j = j + 4
    endfor
    oBloc3 = OBJ_NEW('IDLgrPolygon', bloc3List, $
        POLYGONS=bloc3NewMesh, COLOR=[100,100,255] )
    mainModel->Add, oBloc3

	;
    ;  Create the 4 white line connecting the main satellite
    ;  body to the solar array panels.
    ;
    y = [-0.07, -0.20]
    x = [0.20, 0.38]
    z = [0.0, 0.0]
    oLine1 = OBJ_NEW('IDLgrPolyline', x, y, z, $
        COLOR=[255, 255, 255])
    mainModel->Add, oLine1

    y = [-0.07, -0.20]
    x = [0.20, 0.02]
    z = [0.0, 0.0]
    oLine2 = OBJ_NEW('IDLgrPolyline', x, y, z, $
        COLOR=[255, 255, 255])
    mainModel->Add, oLine2

    y = [0.07, 0.20]
    x = [0.20, 0.38]
    z = [0.0, 0.0]
    oLine3 = OBJ_NEW('IDLgrPolyline', x, y, z, $
        COLOR=[255, 255, 255])
    mainModel->Add, oLine3

    y = [0.07, 0.20]
    x = [0.20, 0.02]
    z = [0.0, 0.0]
    oLine4 = OBJ_NEW('IDLgrPolyline', x, y, z, $
        COLOR=[255, 255, 255])
    mainModel->Add, oLine4

	;
	;	add viewing direction
	;
	x = [0.2, 1.2]
	y = [0, 0]
	z = [0, 0]
	viewing = OBJ_NEW('IDLgrPolyline', x, y, z, COLOR=[255, 0, 0])
	mainModel->Add, viewing

	x = [0.2, 0.2]
	y = [0, 1]
	z = [0, 0]
	viewing = OBJ_NEW('IDLgrPolyline', x, y, z, COLOR=[0, 255, 0])
	mainModel->Add, viewing

	x = [0.2, 0.2]
	y = [0, 0]
	z = [0, 1]
	viewing = OBJ_NEW('IDLgrPolyline', x, y, z, COLOR=[0, 0, 255])
	mainModel->Add, viewing

    mainModel->Scale, 0.5, 0.5, 0.5

    RETURN, mainModel

end


;---------------------------------------------------------------------------
;	Adcance scenery to specific epoch
;---------------------------------------------------------------------------
pro advanceSceneryTo, $
	time, $		;	IN: time to which to advance scenery (MET tag)
	sState		;	IN: state of the application

	;
	;  Increment the time.
	;
	deltaT = time - sState.time
	sState.time = time

	;  Compute new S/C position
	newPos = xmmPosition(sState.time)
	newPosition = newPos/sState.earthRadius
	;print, newPosition
	delta = newPosition - sState.position
	sState.Position = newPosition

	;
	;	Update position labels
	;
	WIDGET_CONTROL, sState.wXLabel, SET_VALUE=$
		STRING(newPos[0], FORMAT='("x=", f11.3, " km")')
	WIDGET_CONTROL, sState.wYLabel, SET_VALUE=$
		STRING(newPos[1], FORMAT='("y=", f11.3, " km")')
	WIDGET_CONTROL, sState.wZLabel, SET_VALUE=$
		STRING(newPos[2], FORMAT='("z=", f11.3, " km")')
	WIDGET_CONTROL, sState.wDLabel, SET_VALUE=$
		STRING(sqrt(newPos[0]*newPos[0]+newPos[1]*newPos[1]+$
				newPos[2]*newPos[2]), $
					  FORMAT='("D=", f11.3, " km")')

	;
	;  Translate the satellite
	;
	sState.oSatModel->Translate, delta[0], delta[1], delta[2]

	;
	;  Rotate the satellite.
	;
	angles = xmmAttitude(sState.time)
	sState.oSatRotationModel->SetProperty, TRANSFORM=sState.initSatTM0
	sState.oSatRotationModel->Rotate, angles[3:5], angles[6]*180./!PI

	;
	;	Update attitude labels
	;
	WIDGET_CONTROL, sState.wRALabel, SET_VALUE=STRING($
		CALL_EXTERNAL('libxmmtrackinfo.so', 'raString', $
		angles[0], RETURN_TYPE=7), FORMAT='("R.A.=", A14)')
	WIDGET_CONTROL, sState.wDECLabel, SET_VALUE=STRING($
		CALL_EXTERNAL('libxmmtrackinfo.so', 'decString', $
		angles[1], RETURN_TYPE=7), FORMAT='("DEC =", A14)')
	WIDGET_CONTROL, sState.wAPOSLabel, SET_VALUE=$
		STRING(angles[2]*180./!DPI, $
		FORMAT='("APOS=", f7.2, " deg")')

	;
	;  Rotate the earth
	;
	earthRotationRate = 0.004178074   ; degrees per sec
	earthAngle = earthRotationRate*deltaT
	sState.oEarthRotationModel->Rotate, [0,0,1], earthAngle

	;
	;	Update time text
	;
	WIDGET_CONTROL, sState.wTimeLabel, SET_VALUE=$
		CALL_EXTERNAL('libxmmtrackinfo.so', 'absTime', $
		sState.time, RETURN_TYPE=7)
	WIDGET_CONTROL, sState.wTimeSlider, SET_VALUE=sState.time

	;
	;  Draw the new frame.
	;
	sState.oWindow->Draw, sState.oView
end


;--------------------------------------------------------------------------
;  Handle all events
;--------------------------------------------------------------------------
pro handleEvents, $
     sEvent       ; IN:  event structure

	;
    ;  Quit the application using the close box.
    ;
    if (TAG_NAMES(sEvent, /STRUCTURE_NAME) EQ $
        'WIDGET_KILL_REQUEST') then begin
        WIDGET_CONTROL, sEvent.top, /DESTROY
        RETURN
    endif

    WIDGET_CONTROL, sEvent.id, GET_UVALUE=uValue

    case uValue of
		;
        ;  Start the animation.
        ;
        'START' : begin
            WIDGET_CONTROL, sEvent.top, GET_UVALUE=sState, /NO_COPY
            WIDGET_CONTROL, sState.wRightBase, TIMER=sState.timer
            sState.stopFlag = 0
            WIDGET_CONTROL, sState.wStopButton, SENSITIVE=1
            WIDGET_CONTROL, sState.wResetButton, SENSITIVE=0
            WIDGET_CONTROL, sState.wStartButton, SENSITIVE=0
            WIDGET_CONTROL, sEvent.top, SET_UVALUE=sState, /NO_COPY
        end

		;
        ;  Handle the widget timer event. This event create another
        ;  widget timer event, hence creating a loop. The loop is
        ;  broken by pushing the 'STOP' button.
        ;
        'TIMER' : begin

            WIDGET_CONTROL, sEvent.top, GET_UVALUE=sState, /NO_COPY

			;
            ;  Stop animation if time has prgressed beyond end time
            ;
            if (sState.time GT sState.endTime) then begin
                WIDGET_CONTROL, sState.wStopButton, SENSITIVE=0
                WIDGET_CONTROL, sState.wResetButton, SENSITIVE=1
                WIDGET_CONTROL, sState.wStartButton, SENSITIVE=0
                WIDGET_CONTROL, sEvent.top, SET_UVALUE=sState, /NO_COPY
                RETURN
            endif

			;
            ;  Process the animation only if the 'STOP' button has
            ;  not been pushed.
            ;
            if (sState.stopFlag EQ 0) then begin
				;
                ;  Increment the time
                ;
				newTime = sState.time + sState.timeIncrement
				advanceSceneryTo, newTime, sState
                sState.time = newTime

				;
                ;  Create a timer event to form the animation loop
                ;
                WIDGET_CONTROL, sState.wRightBase, TIMER=sState.timer

            endif

            WIDGET_CONTROL, sEvent.top, SET_UVALUE=sState, /NO_COPY
         end

		'TIMESCALE' : begin
            WIDGET_CONTROL, sEvent.top, GET_UVALUE=sState, /NO_COPY
			WIDGET_CONTROL, sState.wTimeSlider, GET_VALUE=time
			advanceSceneryTo, time, sState
            WIDGET_CONTROL, sEvent.top, SET_UVALUE=sState, /NO_COPY
		end

		;
        ;  Handle events of the drawing area
        ;
        'DRAWING' : begin

            WIDGET_CONTROL, sEvent.top, GET_UVALUE=sState, /NO_COPY

			;
            ;  Expose
            ;
            if (sEvent.type EQ 4) then begin
                sState.oWindow->draw, sState.oView
                WIDGET_CONTROL, sEvent.top, SET_UVALUE=sState, /NO_COPY
                RETURN
            endif

			;
            ;  Handle trackball update
            ;
            bHaveTransform = sState.oTrack->Update(sEvent, TRANSFORM=qmat )
            if (bHaveTransform NE 0) then begin
                sState.oTopModel->GetProperty, TRANSFORM=t
                mt = t # qmat
                sState.oTopModel->SetProperty,TRANSFORM=mt
            endif

			;
            ;  Handle button press
            ;
            if (sEvent.type EQ 0) then begin
                sState.btndown = 1b
                sState.oWindow->SetProperty, QUALITY=0
                WIDGET_CONTROL,sState.wDraw, /DRAW_MOTION
            endif

			;
            ;  Handle button motion
            ;
            if (sEvent.type EQ 2) then begin
                if (bHaveTransform) then $
                    sState.oWindow->Draw, sState.oView
            endif

			;
            ;  Handle button release
            ;
            if (sEvent.type EQ 1) then begin
                if (sState.btndown EQ 1b) then begin
                    sState.oWindow->SetProperty, QUALITY=2
                    sState.oWindow->Draw, sState.oview
                endif
                sState.btndown = 0b
                WIDGET_CONTROL, sState.wDraw, DRAW_MOTION=0
            endif

            WIDGET_CONTROL, sEvent.top, SET_UVALUE=sState, /NO_COPY
        end

		;
        ;  Stop the animation
        ;
        'STOP' : begin
            WIDGET_CONTROL, sEvent.top, GET_UVALUE=sState, /NO_COPY
            sState.stopFlag = 1
            WIDGET_CONTROL, sState.wStopButton, SENSITIVE=0
            WIDGET_CONTROL, sState.wResetButton, SENSITIVE=1
            WIDGET_CONTROL, sState.wStartButton, SENSITIVE=1
            WIDGET_CONTROL, sEvent.top, SET_UVALUE=sState, /NO_COPY
            RETURN
        end

		;
        ;  Reset the initial state (time and orientation of the objects.)
        ;
        'RESET' : begin
            WIDGET_CONTROL, sEvent.top, GET_UVALUE=sState, /NO_COPY
            WIDGET_CONTROL, sState.wTopBase, SENSITIVE=0
            WIDGET_CONTROL, sState.wStopButton, SENSITIVE=0
            WIDGET_CONTROL, sState.wResetButton, SENSITIVE=0
            WIDGET_CONTROL, sState.wStartButton, SENSITIVE=1

			advanceSceneryTo, sState.startTime, sState
            sState.time = sState.startTime

            WIDGET_CONTROL, sState.wTopBase, SENSITIVE=1
            WIDGET_CONTROL, sEvent.top, SET_UVALUE=sState, /NO_COPY
            RETURN
        end

		;
		;	Add predicted orbit
		;
		'ADDORB' : begin
            WIDGET_CONTROL, sEvent.top, GET_UVALUE=sState, /NO_COPY
			if (sState.wOrbAdded eq 0) then begin
	            WIDGET_CONTROL, sState.wPredOrbTime, GET_VALUE=timeString
				orbEle = DBLARR(3)
				if (CALL_EXTERNAL('libxmmtrackinfo.so', 'xmmPredictedOrbitAt',$
					timeString, orbEle) eq 0) then begin
				;print, orbEle
					WIDGET_CONTROL, sState.wInclLabel, SET_VALUE=$
						STRING(orbEle[0], $
						FORMAT='("Inclination=", f6.1, " deg")')
					WIDGET_CONTROL, sState.wAnodeLabel, SET_VALUE=$
						STRING(orbEle[1], $
						FORMAT='("R.A. a-node=", f6.1, " deg")')
					WIDGET_CONTROL, sState.wAPeriLabel, SET_VALUE=$
						STRING(orbEle[2], $
						FORMAT='("Arg-perigee=", f6.1, " deg")')

					duration = 2.*86400.
					nPoints = fix(duration/100)
					x = fltarr(nPoints)
					y = fltarr(nPoints)
					z = fltarr(nPoints)
				    min = 200000.
					max = 0
					zmin = 10000.
					perdir = DBLARR(3)
					ascdir = DBLARR(2, 3)
					index = 0
					pos0 = xmmPredictedPosition(sState.startTime)/$
							sState.earthRadius
					for i=0, nPoints-1 do begin
						pos1 = xmmPredictedPosition(sState.startTime+$
								i*100.)/sState.earthRadius
						x[i] = pos0[0]
						y[i] = pos0[1]
						z[i] = pos0[2]
						d = sqrt(pos0[0]*pos0[0]+pos0[1]*pos0[1]+$
								 pos0[2]*pos0[2])
						if (d lt min) then begin
							min=d
							perdir = pos0
						end
						if (d gt max) then max=d
						if (pos0[2]*pos1[2] lt 0.) then begin
							;
							;	found orbit node
							;
							ascdir[index, 0:2] = pos0
							index = index + 1
						end
						pos0 = pos1
					end
					;print, min*sState.earthRadius, max*sState.earthRadius
					sState.predictedTrajectory = OBJ_NEW('IDLgrPolyline',$
						 x, y, z, LINESTYLE=0, COLOR=[255, 0,0])
					sState.oTopModel->Add, sState.predictedTrajectory

					sState.perigeeDir = OBJ_NEW('IDLgrPolyline', $
						[0, perdir[0]], [0, perdir[1]], [0, perdir[2]], $
						COLOR=[255, 255, 255])
					sState.ascenDir = OBJ_NEW('IDLgrPolyline', $
								[ascdir[0, 0], ascdir[1, 0]], $
								[ascdir[0, 1], ascdir[1, 1]], $
								[ascdir[0, 2], ascdir[1, 2]], $
								COLOR=[255, 255, 255])

					norma = sqrt(ascdir[0, 0]*ascdir[0, 0]+$
								 ascdir[0, 1]*ascdir[0, 1]+$
								 ascdir[0, 2]*ascdir[0, 2])
					normp = sqrt(perdir[0]*perdir[0]+perdir[1]*perdir[1]+$
									perdir[2]*perdir[2])
					pr = (ascdir[0, 0]*perdir[0]+ascdir[0, 1]*perdir[1]+$
									ascdir[0, 2]*perdir[2])/norma/normp
					;print, "Angle=", acos(pr)*180./!DPI

					sState.oTopModel->Add, sState.perigeeDir
					sState.oTopModel->Add, sState.ascenDir
				    WIDGET_CONTROL, sState.wPredOrbTime, SENSITIVE=0
					WIDGET_CONTROL, sState.wPredOrbAdd, SET_VALUE="Remove"
					sState.wOrbAdded = 1
				end
			end else begin
				;
				;	remove orbit
				;
				sState.oTopModel->Remove, sState.PredictedTrajectory
				sState.oTopModel->Remove, sState.PerigeeDir
				sState.oTopModel->Remove, sState.AscenDir
				sState.wOrbAdded = 0

				WIDGET_CONTROL, sState.wInclLabel, SET_VALUE=$
						"                      "
				WIDGET_CONTROL, sState.wANodeLabel, SET_VALUE=$
						"                      "
				WIDGET_CONTROL, sState.wAPeriLabel, SET_VALUE=$
						"                      "

				WIDGET_CONTROL, sState.wPredOrbTime, SENSITIVE=1
				WIDGET_CONTROL, sState.wPredOrbAdd, SET_VALUE="Add"
			end

            sState.oWindow->Draw, sState.oView

            WIDGET_CONTROL, sEvent.top, SET_UVALUE=sState, /NO_COPY
		end

		;
        ;  Scale the satellite
        ;
        'SCALING' : begin
            WIDGET_CONTROL, sEvent.top, GET_UVALUE=sState, /NO_COPY
            WIDGET_CONTROL, sState.wScalingSlider, GET_VALUE=scale

            scale = 1. + FLOAT(scale) / 100.0
            scalep = scale*100.0
            scalingString = STRING(scalep, FORMAT='(f5.1)') + ' %'
            WIDGET_CONTROL, sState.wScalingLabel2, $
                SET_VALUE=scalingString

            transform = [[scale, 0, 0, 0.0], [0, scale, 0, 0.0], $
            [0, 0, scale, 0.0], [0, 0, 0, 1]]
            sState.oSatScalingModel->SetProperty, TRANSFORM=transform
            WIDGET_CONTROL, sState.wTopBase, /HOURGLASS
            sState.oWindow->Draw, sState.oView
            WIDGET_CONTROL, sEvent.top, SET_UVALUE=sState, /NO_COPY
        end

		;
		;	Zoom into the view
		;
		'ZOOM' : begin
			WIDGET_CONTROL, sEvent.top, GET_UVALUE=sState, /NO_COPY
			WIDGET_CONTROL, sState.wZoomSlider, GET_VALUE=zoom

			f = FLOAT(zoom);
			sState.oView->SetProperty, $
						VIEW=sState.defaultView/f
            WIDGET_CONTROL, sState.wTopBase, /HOURGLASS
            sState.oWindow->Draw, sState.oView
            WIDGET_CONTROL, sEvent.top, SET_UVALUE=sState, /NO_COPY
		end

		;
        ;  Determine the time step between each redraw.
        ;  A larger time step will make the satellite to
        ;  translate at a faster rate but the  animation
        ;  will be less 'smoother'.
        ;
        'STEP' : begin
            WIDGET_CONTROL, sEvent.top, GET_UVALUE=sState, /NO_COPY
            WIDGET_CONTROL, sState.wStepSlider, GET_VALUE=timeIncrement
            sState.timeIncrement = timeIncrement
            WIDGET_CONTROL, sEvent.top, SET_UVALUE=sState, /NO_COPY
        end

		;
        ;  Display the information text file
        ;
        'ABOUT' : BEGIN
        END   ; of ABOUT

		;
        ;  Quit this application
        ;
        'QUIT' : begin
            WIDGET_CONTROL, sEvent.top, /DESTROY
			EXIT
        end

        ELSE :   ;  do nothing

    endcase
end


;--------------------------------------------------------------------------
;  Destroy the top objects and restore the previous color table
;--------------------------------------------------------------------------
pro cleanup, $
    wTopBase        ;  IN: top level base identifier

    WIDGET_CONTROL, wTopBase, GET_UVALUE=sState, /NO_COPY

    ;  Destroy the top objects
    ;
    OBJ_DESTROY, sState.oView
    OBJ_DESTROY, sState.oTrack
    OBJ_DESTROY, sState.oText
    OBJ_DESTROY, sState.oFont

    ;  Restore the color table
    ;
    TVLCT, sState.colorTable

    if WIDGET_INFO(sState.groupBase, /VALID_ID) then $
        WIDGET_CONTROL, sState.groupBase, /MAP

end

;--------------------------------------------------------------------------
;  Show XMM orbiting the Earth
;--------------------------------------------------------------------------
pro xmmtrack

	;
	;	obtain start/stop times of observation
	;
	startTime = CALL_EXTERNAL('libxmmtrackinfo.so', 'startTime', RETURN_TYPE=5)
	;	print, startTime
	endTime = CALL_EXTERNAL('libxmmtrackinfo.so', 'endTime', RETURN_TYPE=5)
	;	print, endTime


    groupBase = 0L

	;
    ;  Get the current color vectors to restore
    ;  when this application is exited.
	;
    TVLCT, savedR, savedG, savedB, /GET

	;
    ; Build color table from color vectors
    ;
    colorTable = [[savedR],[savedG],[savedB]]

	;
    ;  Set up dimensions of the drawing (viewing) area
    ;
    DEVICE, GET_SCREEN_SIZE=screenSize
    xdim = screenSize[0]*0.6
    ydim = xdim*0.8

	;
    ;  Create widgets
    ;
	title="XMM's trajectory in revolution " + $
			strcompress(string(xmmRevolutionNr()), /REMOVE_ALL)
    wTopBase = WIDGET_BASE(/COLUMN, $
            TLB_FRAME_ATTR=1, MBAR=barBase, $
            XPAD=0, YPAD=0, $
            /TLB_KILL_REQUEST_EVENTS, $
            TITLE=title)

	;
    ;  Create the menu bar. It contains the file,
    ;  edit, and help buttons.
    ;
    wFileButton = WIDGET_BUTTON(barBase, VALUE = 'File', /MENU)

    wQuitButton = WIDGET_BUTTON(wFileButton, VALUE='Quit', $
            UVALUE='QUIT')

	;
    ;  Create the menu bar item help that contains the about button
    ;
    wHelpMenu = WIDGET_BUTTON(barBase, VALUE='About', $
        /HELP, /MENU)

    wAboutButton = WIDGET_BUTTON(wHelpMenu, $
            VALUE='XMM Track', UVALUE='ABOUT')

	;
    ;  Create a sub base of the top base (wBase)
    ;
    wSubBase = WIDGET_BASE(wTopBase, COLUMN=2)

	;
    ;  Create the left Base that contains the functionality buttons.
    ;  Here the only button is to animate the object.
    ;
    wLeftbase = WIDGET_BASE(wSubBase,/BASE_ALIGN_CENTER, COLUMN=1)

    wAnimationBase = WIDGET_BASE(wLeftBase, /COLUMN, YPAD=5)
    wAnimationLabel = WIDGET_LABEL(wAnimationBase, VALUE='Animation')
    wStartButton = WIDGET_BUTTON(wAnimationBase, VALUE="Start", UVALUE='START')
    wStopButton = WIDGET_BUTTON(wAnimationBase, VALUE="Stop", UVALUE='STOP')
    wResetButton = WIDGET_BUTTON(wAnimationBase, VALUE="Reset", UVALUE='RESET')

    wScalingBase = WIDGET_BASE(wLeftBase, /COLUMN, YPAD=5)
    wScalingLabel1 = WIDGET_LABEL(wScalingBase, VALUE='Satellite Scaling')

    percent = 100
    scalingString = STRING(percent, FORMAT='(f5.1)') + ' %'
    wScalingLabel2 = WIDGET_LABEL(wScalingBase, VALUE=scalingString)
    wScalingSlider = WIDGET_SLIDER(wScalingBase, $
                    MINIMUM=0, $
                    MAXIMUM=800, VALUE=0, $
                    /SUPPRESS_VALUE, $
                    UVALUE='SCALING')

	wZoomBase = WIDGET_BASE(wLeftBase, /COLUMN, YPAD=5)
	wZoomLabel = WIDGET_LABEL(wZoomBase, VALUE='Zoom view')
	wZoomSlider = WIDGET_SLIDER(wZoomBase, $
					MINIMUM=1, $
					MAXIMUM=10, $
					VALUE=2, $
					UVALUE='ZOOM')

    wStepBase = WIDGET_BASE(wLeftBase, /COLUMN, YPAD=5)
    wStepLabel = WIDGET_LABEL(wStepBase, VALUE='Animation Step [s]')
    wStepSlider = WIDGET_SLIDER(wStepBase, $
                    MINIMUM=100, $
                    MAXIMUM=5000, VALUE=100, $
                    UVALUE='STEP')

	startPos = xmmPosition(startTime)
	wPosBase = WIDGET_BASE(wLeftBase, /COLUMN, YPAD=10)
	wxLabel = WIDGET_LABEL(wPosBase, $
				VALUE=STRING(startPos[0], FORMAT='("x=", f11.3, " km")'))
	wyLabel = WIDGET_LABEL(wPosBase, $
				VALUE=STRING(startPos[1], FORMAT='("y=", f11.3, " km")'))
	wzLabel = WIDGET_LABEL(wPosBase, $
				VALUE=STRING(startPos[2], FORMAT='("z=", f11.3, " km")'))
	wdLabel = WIDGET_LABEL(wPosBase, $
				VALUE=STRING(sqrt(startPos[0]*startPos[0]+$
								  startPos[1]*startPos[1]+$
								  startPos[2]*startPos[2]), $
									  FORMAT='("D=", f11.3, " km")'))

	startAtt = xmmAttitude(startTime)
	wAttBase = WIDGET_BASE(wLeftBase, /COLUMN, YPAD=10)
	wRALabel = WIDGET_LABEL(wAttBase, /ALIGN_LEFT, VALUE=STRING($
					CALL_EXTERNAL('libxmmtrackinfo.so', 'raString', $
					startAtt[0], RETURN_TYPE=7), FORMAT='("R.A.=", A14)'))
	wDECLabel = WIDGET_LABEL(wAttBase, /ALIGN_LEFT, VALUE=STRING($
					CALL_EXTERNAL('libxmmtrackinfo.so', 'decString', $
					startAtt[1], RETURN_TYPE=7), FORMAT='("DEC =", A14)'))
	wAPOSLabel = WIDGET_LABEL(wAttBase, /ALIGN_LEFT, VALUE=STRING($
					startAtt[2]*180./!DPI, FORMAT='("APOS=", f7.2, " deg")'))

	wPredOrbBase = WIDGET_BASE(wLeftBase, /COLUMN, YPAD=5)
	wPredOrbLabel = WIDGET_LABEL(wPredOrbBase, VALUE='Predicted orbit at')

	textStartTime = CALL_EXTERNAL('libxmmtrackinfo.so', 'absTime', $
			startTime, RETURN_TYPE=7)
	wPredOrbTime = WIDGET_TEXT(wPredOrbBase, VALUE=textStartTime, /EDITABLE)
    wPredOrbAdd = WIDGET_BUTTON(wPredOrbBase, VALUE="Add", UVALUE='ADDORB')
	wInclLabel = WIDGET_LABEL(wPredOrbBase, /ALIGN_LEFT, $
		VALUE="                      ")
	wANodeLabel = WIDGET_LABEL(wPredOrbBase, /ALIGN_LEFT, $
		VALUE="                      ")
	wAPeriLabel = WIDGET_LABEL(wPredOrbBase, /ALIGN_LEFT, $
		VALUE="                      ")

	;
    ;  Create the right Base that has the drawing area.
    ;
    wRightbase = WIDGET_BASE(wSubBase, COLUMN=1, UVALUE='TIMER')
    wDraw = WIDGET_DRAW(wRightBase, $
                GRAPHICS_LEVEL=2, $
                XSIZE=xdim, YSIZE=ydim, /BUTTON_EVENTS, $
                UVALUE='DRAWING', RETAIN=0, /EXPOSE_EVENT)
	wTimeBase = WIDGET_BASE(wRightBase, /COLUMN, YPAD=5, XSIZE=xdim)
	wTimeLabel = WIDGET_LABEL(wTimeBase, VALUE=textStartTime)
	wTimeSlider = WIDGET_SLIDER(wTimeBase, $
					MINIMUM=startTime, $
					MAXIMUM=endTime, $
					VALUE=startTime, $
					/SUPPRESS_VALUE, UVALUE='TIMESCALE', XSIZE=xdim)

    wStatusBase = WIDGET_BASE(wTopBase, MAP=0, /ROW)

	;
	;  Now the widget have been created, realize it.
    ;
    WIDGET_CONTROL, wTopBase, /REALIZE

	;
    ; Returns the top level base to the APPTLB keyword
    ;
    appTLB = wtopBase

    WIDGET_CONTROL, wTopBase, SENSITIVE=0

	;
    ;  Grab the window id of the drawable
    ;
    WIDGET_CONTROL, wDraw, GET_VALUE=oWindow

	;
    ;  Compute viewplane rectangle based on aspect ratio
    ;
    aspect = float(xdim)/float(ydim)
    defaultView = [-20,-20,40,40]
	defaultEye = 50
    if (aspect > 1) then begin
        defaultView[0] = defaultView[0] - ((aspect-1.0)*defaultView[2])/2.0
        defaultView[2] = defaultView[2] * aspect
    endif else begin
        defaultView[1] = $
					defaultView[1] - (((1.0/aspect)-1.0)*defaultView[3])/2.0
        defaultView[3] = defaultView[3] * aspect
    endelse

	;
    ; Create view object
    ;
    oView = OBJ_NEW('idlgrview', PROJECTION=2, EYE=defaultEye, $
        ZCLIP=[20,-100.0], VIEW=.5*defaultView, COLOR=[0, 0, 0])

	;
    ;  Create a centered starting up text
    ;
    textLocation = [defaultView[0]+0.5*defaultView[2], $
					defaultView[1]+0.5*defaultView[3]]

	;
    ;  Create and display the PLEASE WAIT text.
    ;
    oFont = OBJ_NEW('IDLgrFont', 'Helvetica', SIZE=12)
    oText = OBJ_NEW('IDLgrText', $
        'Starting up  Please wait...', $
        ALIGN=0.5, $
        LOCATION=textLocation, $
        COLOR=[255,255,0], FONT=oFont)

	;
    ; Create models and its tree structure
    ;
    oTopModel = OBJ_NEW('idlgrmodel')
        oEarthModel = OBJ_NEW('idlgrmodel')
        oEarthRotationModel = OBJ_NEW('idlgrmodel')
        oEarthScalingModel = OBJ_NEW('idlgrmodel')
        oSatModel = OBJ_NEW('idlgrmodel')
        oSatRotationModel = OBJ_NEW('idlgrmodel')
        oSatScalingModel = OBJ_NEW('idlgrmodel')

	oTextModel = OBJ_NEW('idlgrmodel')
    oTextModel->Add, oText

	;
    ;  Place the model in the view
    ;
    oView->Add, oTopModel
    oView->Add, oTextModel

    oWindow->Draw, oView

    oTopModel->Add, oEarthModel
    oTopModel->Add, oSatModel
    oEarthModel->Add, oEarthRotationModel
    oEarthRotationModel->Add, oEarthScalingModel

    ra0line = OBJ_NEW('IDLgrPolyline', [0, 10], [0, 0], [0, 0], $
        COLOR=[255, 0, 0])
    ra90line = OBJ_NEW('IDLgrPolyline', [0, 0], [0, 10], [0, 0], $
        COLOR=[0, 255, 0])
    dec90line = OBJ_NEW('IDLgrPolyline', [0, 0], [0, 0], [0, 10], $
        COLOR=[0, 0, 255])
	oEarthModel->Add, ra0line
	oEarthModel->Add, ra90line
	oEarthModel->Add, dec90line

    oSatModel->Add, oSatRotationModel
    oSatRotationModel->Add, oSatScalingModel
    scale = 2.0
    oEarthScalingModel->Scale, scale, scale, scale

	;
    ;  Scale the top model to fit the viewing area
    ;
    sct = 0.7
    oTopModel->Scale, sct, sct, sct
	oTopModel->Rotate, [1, 0, 0], -90

	;
    ;  Create lights
    ;
	;	first create dim ambient light source
	;
    oLIght1 = OBJ_NEW('IDLgrLight', TYPE=0, INTENSITY=.2)
    oTopModel->Add, oLight1

	;
	;	Sun: try to compute position, if that fails, position the sun
	;		 at Ra=12h, Dec=0, i.e., it shines in dirction [1, 0, 0]
	;
	earthSunVec = DBLARR(3)
	if (CALL_EXTERNAL('libxmmtrackinfo.so', 'sunPosition', $
		startTime, earthSunVec) ne 0) $
		then earthSunVec = [-1, 0, 0]
	;print, earthSunVec
    oLIght2 = OBJ_NEW('IDLgrLight', TYPE=2, INTENSITY=2., $
		LOCATION=10*earthSunVec, DIRECTION=-earthSunVec)
    oTopModel->Add, oLight2

	;
    ;  Create the earth
    ;
    READ_JPEG, getenv('XMMTRACK_DIR') + "earth.jpg", image, TRUE=1
    oImageArray = OBJ_NEW('IDLgrImage', image, HIDE=1)
    oEarthScalingModel->Add, oImageArray

    oPlanet = OBJ_NEW('orb', COLOR=[255, 255, 255], RADIUS=.5, $
        DENSITY=0.8, /TEX_COORDS, TEXTURE_MAP=oImageArray)

    oEarthScalingModel->Add, oPlanet

	;
	;	adjust initial earth rotation angle:
	;	1. rotate about 180 degrees because left edge of earth texture
	;	   map is at longitude=12h
	;	2. rotate about Greewich sidereal time at startTime
	;
	siderealTime = CALL_EXTERNAL('libxmmtrackinfo.so', 'siderealTime', $
		startTime, RETURN_TYPE=5) * 180./!DPI
	;print, siderealTime

	oEarthRotationModel->Rotate, [0, 0, 1], 180 + siderealTime
    oEarthRotationModel->GetProperty, TRANSFORM=initEarthTM
    oSatRotationModel->GetProperty, TRANSFORM=initSatTM0
    oTopModel->GetProperty, TRANSFORM=initTopTM

	;
    ;  Create a satellite object placed into a model,
    ;  return that model.
    ;
    oSatelliteModel = xmmCreateModel()
    oSatScalingModel->Add, oSatelliteModel

    ;  Translate the satellite and place it to its
    ;  location at the epoch time
    ;
    position = xmmPosition(startTime)
	earthRadius = 6378.137
    newPosition = position/earthRadius
    oSatModel->Translate, newPosition[0], newPosition[1], newPosition[2]

	;
	;	draw trajectory as polyline in steps of 100s
	;
	duration = endTime-StartTime;
	nPoints = fix(duration/100)
	x = fltarr(nPoints)
	y = fltarr(nPoints)
	z = fltarr(nPoints)
	for i=0, nPoints-1 do begin
		pos = xmmPosition(startTime + i*100.)/earthRadius
		x[i] = pos[0]
		y[i] = pos[1]
		z[i] = pos[2]
	end
	actualTrajectory = OBJ_NEW('IDLgrPolyline', x, y, z, COLOR=[255, 255, 0])
	oTopModel->Add, actualTrajectory

	;
    ;  Add the trackball object for interactive change
    ;  of the scene orientation
    ;
    oTrack = OBJ_NEW('Trackball', [xdim/2.0, ydim/2.0], xdim/2.0)

    sState = { $
        ColorTable: colorTable, $            ; Color table to restore
        Btndown:0b, $                        ; 0= not presses, pressed otherwise
        InitEarthTM: initEarthTM, $          ; Initial Orientation of Earth
        InitSatTM: initSatTM0, $             ; Initial Satellite transformation
        InitSatTM0: initSatTM0, $            ; Initial Satellite transformation
        InitTopTM: initTopTM, $              ; Initial transformationo of top model
        InitSatPosition: newPosition, $      ; Initial Position of the satellite
        StopFlag: 0, $                       ; 0= animation not stopped, 1=stopped
        Position: newPosition, $             ; Updated position of the satellite
        Timer: 0.05, $                        ; Time delay between frame
		StartTime: startTime, $				 ; Start of observation
		EndTime: endTime, $				     ; End of observation
        Time: startTime, $		             ; Accumulated time (sec.)
        TimeIncrement: 100, $                ; Time between frames (sec.)
		EarthRadius: earthRadius, $			 ; radius of the blue planet
        WTopBase: wTopBase, $                ; Top level base IDs
        WRightBase: wRightBase, $            ; Right base ID
        WStartButton: wStartButton, $        ; Animation buttons IDs
        WResetButton: wResetButton, $
        WStopButton: wStopButton, $
        WScalingSlider: wScalingSlider, $    ; Sliders IDs
        WScalingLabel2: wScalingLabel2, $    ; Sliders IDs
		WZoomSlider: wZoomSlider, $			 ; Zoom Slider
        WStepSlider: wStepSlider, $
		WTimeLabel: wTimeLabel, $			 ; label with time text
		WTimeSlider: wTimeSlider, $
		WxLabel: wxLabel, $
		WyLabel: wyLabel, $
		WzLabel: wzLabel, $
		WdLabel: wdLabel, $
		WRALabel: wRALabel, $
		WDECLabel: wDECLabel, $
		WAPOSLabel: wAPOSLabel, $
		WPredOrbTime: wPredOrbTime, $
		WPredOrbAdd: wPredOrbAdd, $
		WOrbAdded: 0, $
		WInclLabel: wInclLabel, $
		WANodeLabel: wANodeLabel, $
		WAPeriLabel: wAPeriLabel, $
		PredictedTrajectory: OBJ_NEW('IDLgrPolyline'), $
		PerigeeDir: OBJ_NEW('IDLgrPolyline'), $
		AscenDir: OBJ_NEW('IDLgrPolyline'), $
        WDraw: wDraw, $                      ; Widget draw ID
        OView: oView, $                      ; View object
        OWindow: oWindow, $                  ; Window object
        OSatModel: oSatModel, $              ; Satellite models
        OSatRotationModel: oSatRotationModel, $
        OSatScalingModel: oSatScalingModel, $
        OEarthRotationModel: oEarthRotationModel, $ ; Earth models
        OTrack: oTrack, $                    ; Trackball object
        OFont: oFont, $                      ; Font object
        OText: oText, $                      ; Text object
		OTextModel: oTextModel, $			 ; Text model
        OTopModel:oTopModel, $               ; Top model
        groupBase: groupBase, $              ; Base of Group Leader
		DefaultView: defaultView, $			 ; default viewing rectangle
		DefaultEye: defaultEye $			 ; default eye distance
    }     

    WIDGET_CONTROL, wTopBase, SET_UVALUE=sState, /NO_COPY
    WIDGET_CONTROL, wStartButton, SENSITIVE=0
    WIDGET_CONTROL, wResetButton, SENSITIVE=0

	;
    ;  Draw the first view.
    ;
    WIDGET_CONTROL, wTopBase, /HOURGLASS
    oTextModel->Remove, oText
    oWindow->Draw, oView

    WIDGET_CONTROL, wTopBase, SENSITIVE=1

    ;  Animate at start up.
    ;
    pseudoEvent = { $
       ID: wRightBase, $
       TOP: wTopBase, $
       HANDLER: wTopBase $
    }

    handleEvents, pseudoEvent

    XMANAGER, 'xmmTrack', wTopBase, EVENT_HANDLER='handleEvents', $
        /NO_BLOCK, CLEANUP='cleanup'

end
